@@ -2,7 +2,7 @@
    daemon.c
 
    This file is part of GNU Anubis.
-   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2007 The Anubis Team.
+   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2007, 2009 The Anubis Team.
 
    GNU Anubis is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -28,6 +28,9 @@ int deny_severity = LOG_INFO;
 int allow_severity = LOG_INFO;
 #endif /* USE_LIBWRAP */
 
+char *log_tag = "anubis";
+int log_facility = LOG_MAIL;
+
 /************
   DAEMONIZE
 *************/
@@ -63,10 +66,8 @@ daemonize (void)
   topt &= ~T_FOREGROUND;
   topt |= T_DAEMON;
 
-#ifdef HAVE_SYSLOG
-  openlog ("anubis", LOG_PID, 0);
+  openlog (log_tag, LOG_PID, log_facility);
   info (NORMAL, _("%s daemon startup succeeded."), version);
-#endif /* HAVE_SYSLOG */
   write_pid_file ();
   return;
 }
@@ -179,22 +180,25 @@ anubis_child_main (struct sockaddr_in *addr)
   int rc;
 
   proclist_init ();
-#ifdef WITH_GSASL
   switch (anubis_mode)
     {
     case anubis_transparent:
       rc = anubis_transparent_mode (addr);
       break;
 
+#ifdef WITH_GSASL
     case anubis_authenticate:
       rc = anubis_authenticate_mode (addr);
+      break;
+#endif /* WITH_GSASL */
 
+    case anubis_proxy:
+      rc = anubis_proxy_mode (addr);
+      break;
+      
     default:
       abort();
     }
-#else
-  rc = anubis_transparent_mode (addr);
-#endif /* WITH_GSASL */
   proclist_cleanup (subprocess_report_status);
   net_close_stream (&remote_client);
   return rc;
@@ -373,12 +377,7 @@ stdinout (void)
   anubis_getlogin (&session.clientname);
   auth_tunnel ();		/* session.clientname = session.supervisor */
 
-  if (!(topt & T_LOCAL_MTA) && !session.mta)
-    {
-      options.termlevel = NORMAL;
-      anubis_error (EXIT_FAILURE, 0, _("The MTA has not been specified. "
-			               "Set the REMOTE-MTA or LOCAL-MTA."));
-    }
+  ASSERT_MTA_CONFIG ();
 
   create_stdio_stream (&remote_client);
 
